
# Compact Crypto Certs (C3) Public API

__version__ = "0.9.1"


from c3.constants import *
from c3.errors import *
from c3 import signverify
from c3 import structure
from c3 import textfiles

# quick-start APIs for verifying
from c3.textfiles import pub_block_from_string
from c3.signverify import SignVerify

# "note: __all__ affects the from <module> import * behavior only."
# In theory these two things are all that is needed to code up a verifier.
__all__ = [
    "pub_block_from_string",
    "SignVerify"
]


# Future todos:
# - py2 support (b3 utf8, base64 encodebytes, p_p pathlib2, more?)
# - more key types e.g. libsodium





# C3 data schema for the binary blocks, tag constants, and API enums

import b3
b3.composite_schema.strict_mode = True

# --- API Actions ---

# Make/Sign actions:
MAKE_SELFSIGNED = 1
MAKE_INTERMEDIATE = 2
SIGN_PAYLOAD = 3

LINK_APPEND = 1
LINK_NAME = 2

# --- environment variables for priv key crypting ---
PASS_VAR = "C3_PASSWORD"
SHOW_VAR = "C3_SHOW_PASS"


# --- Schemas and header tag values ---

# Tag/Key values
# Public-part top level
KEY_LIST_PAYLOAD = 55  # cert chain with a payload as the first entry
KEY_LIST_CERTS = 66   # cert chain with a cert as the first entry

# Public-part chain-level
KEY_DAS = 77            # "data_part and sig_part structure"

# Private-part top level
KEY_PRIV_CRCWRAPPED = 88       # "priv data with a crc32 integrity check"

# Private-part field types
PRIVTYPE_BARE = 1
PRIVTYPE_PASS_PROTECT = 2
KEYTYPE_ECDSA_256P = 1      # this may include hashers (tho may include hashtype later?)
KNOWN_PRIVTYPES = [1, 2]
KNOWN_KEYTYPES = [1]

# --- Data structures ---

CERT_SCHEMA = (
    (b3.BYTES,     "cert_id",       0, True),
    (b3.UTF8,      "subject_name",  1, True),
    (b3.UVARINT,   "key_type",      2, True),
    (b3.BYTES,     "public_key",    3, True),
    (b3.BASICDATE, "expiry_date",   4, True),
    (b3.BASICDATE, "issued_date",   5, True),
)

SIG_SCHEMA = (
    (b3.BYTES, "signature", 0,  True),
    (b3.BYTES, "signing_cert_id", 1, False),  # value can be empty.
)

DATA_AND_SIG = (
    (b3.BYTES, "data_part", 0, True),  # a cert (CERT_SCHEMA) or a payload (BYTES)
    (b3.BYTES, "sig_part", 1, True),   # a SIG_SCHEMA
    # (We could put a sig_list item here later if we want to go chain multi sig.)
)

# --- Private structure stuff ---

PRIV_CRCWRAPPED = (
    (b3.UVARINT, "priv_type", 0, True),      # protection method (e.g. bare/none, or pass_protect)
    (b3.UVARINT, "key_type",  1, True),      # actual type of private key (e.g. ecdsa 256p)
    (b3.BYTES,   "priv_data", 2, True),
    (b3.UVARINT, "crc32",    3, True),       # crc of privdata for integrity check
)


KEY2NAME = {55 : "KEY_LIST_PAYLOAD", 66 : "KEY_LIST_CERTS", 77 : "KEY_DAS", 88 : "KEY_PRIV_CRCWRAPPED"}




# C3 Signing and verifying actions, and a little in-ram trust store for verify().

import datetime

import ecdsa

from c3.constants import *
from c3.errors import *
from c3 import structure
from c3 import getpassword
from c3 import pass_protect
from c3.structure import AttrDict

# Policy: verify() only reads from self.trusted_certs, it doesnt write anything into there.
#         Caller/User must call add_trusted_certs() to add theirs.

# Note: the priv key encrypt/decrypt functions are here too, just to keep things to 1 class for now.

class SignVerify(object):
    def __init__(self):
        self.trusted_certs = {}   # by name. For e.g. root certs etc.

        # We need to load libsodium on startup if it is there. But also support cases where it's
        # not available (e.g.bare keys)
        self.load_error = None
        self.pass_protect = None
        try:
            self.pass_protect = pass_protect.PassProtect()
        except OSError as e:         # No libsodium dll found
            self.load_error = "Starting pass_protect: "+str(e)

    # ============ Verify ==========================================================================

    # In: Data-And-Sigs items list
    # Out: payload bytes or an Exception

    # Apart from actual signature fails, there are 3 other ways for this to fail:
    # 1) unnamed issuer cert and no next cert in line aka "fell off the end" (ShortChainError)
    # 2) Named cert not found - in the cert store / trust store / certs_by_name etc
    # 3) Last cert is self-signed and verifies OK but isn't in the trust store. (Untrusted Chain)

    def verify(self, chain):
        certs_by_id = {das.cert.cert_id : das.cert for das in chain if "cert" in das}
        found_in_trusted = False         # whether we have established a link to the trusted_certs

        for i, das in enumerate(chain):
            signing_cert_id = das.sig.signing_cert_id
            # --- Find the 'next cert' ie the one which verifies our signature ---
            if not signing_cert_id:
                # --- no signing-id means "next cert in the chain" ---
                if i + 1 >= len(chain):      # have we fallen off the end?
                    raise ShortChainError(structure.ctnm(das)+"Next issuer cert is missing")
                next_cert = chain[i + 1].cert
            else:
                # --- got a name, look in trusted for it, then in ourself (self-signed) ---
                if signing_cert_id in self.trusted_certs:
                    next_cert = self.trusted_certs[signing_cert_id]
                    found_in_trusted = True
                elif signing_cert_id in certs_by_id:
                    next_cert = certs_by_id[signing_cert_id]
                else:
                    raise CertNotFoundError(structure.ctnm(das)+"wanted cert %r not found" % signing_cert_id)

            # --- Actually verify the signature ---
            try:
                VK = ecdsa.VerifyingKey.from_string(next_cert.public_key, ecdsa.NIST256p)
                VK.verify(das.sig.signature, das.data_part)  # returns True or raises exception
            except Exception:       # wrap theirs with our own error class
                raise InvalidSignatureError(structure.ctnm(das)+"Signature failed to verify")
            # --- Now do next das in line ---

        # Chain verifies completed without problems. Make sure we got to a trust store cert.
        # If there is a payload, return it otherwise return True. (So bool-ness works.)
        # Even tho all errors are exceptions.
        if found_in_trusted:
            return True
        raise UntrustedChainError("Chain does not link to trusted certs")

    # --- in-ram trust store for Verify's trust anchors. ---

    def add_trusted_certs(self, certs_bytes, force=False):
        cert_chain = structure.load_pub_block(certs_bytes)
        if not force:
            try:
                self.verify(cert_chain)
            except UntrustedChainError:   # ignore this one failure mode because we havent installed
                pass                      # this/these certs yet!
        for das in cert_chain:
            if "cert" not in das:         # skip payload if there is one
                continue
            self.trusted_certs[das.cert.cert_id] = das.cert
        return


    # ============ Signing =========================================================================

    #               |     no payload             payload
    #  -------------+-------------------------------------------------
    #  using cert   |     make chain signer      sign payload
    #               |
    #  using self   |     make self signer       ERROR invalid state

    # The way sign combines things, we get bytes out.
    # Cannot be seperated like load and verify are, nor do we want to.

    # In: a cert name and/or payload, and a cert/keypair to use to sign with (if applicable)
    # Out: public_part BLOCK, private key BYTES
    # Note: incoming and outgoing privs are vare, caller must encrypt/decrypt.

    def make_sign(self, action, name="", payload=b"", using_priv=b"", using_pub=b"", using_name="", expiry=None, link=LINK_APPEND):
        # using_pub must now always be present unless MAKE_SELFSIGNED
        if action != MAKE_SELFSIGNED:
            if not using_pub:
                raise ValueError("please supply public part of --using")
            # Make sure the signing cert hasn't expired!
            structure.ensure_not_expired(using_pub)
            # Make sure the keypair really is a keypair
            self.check_privpub_match_ecdsanist256p(using_priv, using_pub)

        # sanity check expiry, needed for selfsign and intermediate
        if action in (MAKE_INTERMEDIATE, MAKE_SELFSIGNED):
            if not expiry:
                raise ValueError("creating cert: please provide expiry date")

        # Policy: for now, cert_id and subject name are the same.  Consider ULID in future.
        cert_id = name.encode("ascii")
        using_id = using_name.encode("ascii")
        new_key_priv = None

        # --- Key gen if applicable ---
        if action in (MAKE_SELFSIGNED, MAKE_INTERMEDIATE):
            # make keys
            new_key_priv, new_key_pub = self.gen_keys_ECDSA_nist256p()
            # make pub cert for pub key

            today = datetime.date.today()
            new_pub_cert = AttrDict(
                public_key=new_key_pub, subject_name=name, cert_id=cert_id, issued_date=today,
                key_type=KEYTYPE_ECDSA_256P, expiry_date=expiry
            )
            from pprint import pprint
            pprint(new_pub_cert)
            new_pub_cert_bytes = b3.schema_pack(CERT_SCHEMA, new_pub_cert)
            payload_bytes = new_pub_cert_bytes
        # --- Load payload if not key-genning ---
        else:
            payload_bytes = payload

        if action == MAKE_SELFSIGNED:
            # --- Sign the cert using its own private key ---
            SK = ecdsa.SigningKey.from_string(new_key_priv, ecdsa.NIST256p)
            sig_d = AttrDict(signature=SK.sign(payload_bytes), signing_cert_id=cert_id)

        else:
            # --- Sign the thing (cert or payload) using Using, if not selfsign ----
            SK = ecdsa.SigningKey.from_string(using_priv, ecdsa.NIST256p)
            sig_d = AttrDict(signature=SK.sign(payload_bytes), signing_cert_id=using_id)
            # note  ^^^ signing_cert_id can be blank, in which case using_pub should have bytes to append

        sig_part = b3.schema_pack(SIG_SCHEMA, sig_d)

        # --- Make data-and-sig structure ---
        das = AttrDict(data_part=payload_bytes, sig_part=sig_part)
        das_bytes = b3.schema_pack(DATA_AND_SIG, das)

        # --- prepend header for das itself so straight concatenation makes a list-of-das ---
        das_bytes_with_hdr = b3.encode_item_joined(KEY_DAS, b3.DICT, das_bytes)

        # --- Append Using's public_part (the chain) if applicable ---
        if link == LINK_APPEND and action != MAKE_SELFSIGNED:
            # we need to:
            # 1) strip using_public_part's public_part header,
            # (This should also ensure someone doesn't try to use a payload cert chain instead of a signer cert chain to sign things)
            _, index = structure.expect_key_header([KEY_LIST_CERTS], b3.LIST, using_pub, 0)
            using_pub = using_pub[index:]
            # 2) concat our data + using_public_part's data
            out_public_part = das_bytes_with_hdr + using_pub
        else:
            out_public_part = das_bytes_with_hdr

        # --- Prepend a new overall public_part header & return pub & private bytes ---
        if action == SIGN_PAYLOAD:
            key_type = KEY_LIST_PAYLOAD
        else:
            key_type = KEY_LIST_CERTS
        out_public_with_hdr = b3.encode_item_joined(key_type, b3.LIST, out_public_part)

        return out_public_with_hdr, new_key_priv


    # In: nothing
    # Out: key pair as priv bytes and pub bytes

    def gen_keys_ECDSA_nist256p(self):
        curve = [i for i in ecdsa.curves.curves if i.name == 'NIST256p'][0]
        priv = ecdsa.SigningKey.generate(curve=curve)
        pub = priv.get_verifying_key()
        return priv.to_string(), pub.to_string()

    # In: priv key bytes, pub chain block
    # Out: true or exception

    def check_privpub_match_ecdsanist256p(self, priv_key_bytes, using_pub_block):
        using_pub_key = structure.extract_first_dict(using_pub_block, CERT_SCHEMA).public_key
        priv = ecdsa.SigningKey.from_string(priv_key_bytes, ecdsa.NIST256p)
        pub = priv.get_verifying_key()
        if pub.to_string() != using_pub_key:
            raise SignError("private key and public key do not match")
        return True


    # ============== Private key encrypt/decrypt ===================================================

    # NOTE that these interact with the user (password entry)
    #         and use a pass_protect object which needs startup initialisation (load libsodium)

    # In:  private key bytes  (and possibly user entering a password interactively)
    # Out: encrypted private key bytes

    def encrypt_private_key(self, priv_bytes):
        if not self.pass_protect:
            raise RuntimeError(self.load_error)
        prompt1 = "Private  key encryption password: "
        prompt2 = "Re-enter key encryption password: "
        passw = getpassword.get_double_enter_setting_password(prompt1, prompt2)
        if not passw:
            raise ValueError("No password supplied, exiting")
        epriv_bytes = self.pass_protect.SinglePassEncrypt(priv_bytes, passw)
        return epriv_bytes

    # In: dict from load_priv_block
    # Out: private key bytes for make_sign to use

    def decrypt_private_key(self, privd):
        if privd.priv_type == PRIVTYPE_BARE:
            return privd.priv_data
        if privd.priv_type != PRIVTYPE_PASS_PROTECT:
            raise StructureError \
                ("Unknown privtype %d in priv block (wanted %r)" % (privd.priv_type, KNOWN_PRIVTYPES))
        if not self.pass_protect:       # usually because could not find libsodium DLL
            raise RuntimeError(self.load_error)
        if self.pass_protect.DualPasswordsNeeded \
                (privd.priv_data):  # todo: we dont support this here yet
            raise NotImplementedError("Private key wants dual passwords")

        # --- Try password from environment variables ---
        passw = getpassword.get_env_password()
        if passw:
            priv_ret = self.pass_protect.SinglePassDecrypt(privd.priv_data, passw)
            # Note exceptions are propagated right out here, its an exit if this decrypt fails.
            return priv_ret

        # --- Try password from user ---
        prompt = "Password to unlock private key: "
        priv_ret = b""
        while not priv_ret:
            passw = getpassword.get_enter_password(prompt)
            if not passw:
                raise ValueError("No password supplied, exiting")

            try:
                priv_ret = self.pass_protect.SinglePassDecrypt(privd.priv_data, passw)
            except Exception as e:
                print("Failed decrypting private key: ", str(e))
                continue        # let user try again
        return priv_ret





from __future__ import unicode_literals     # for python2

import base64, traceback, random, os, datetime
from pprint import pprint

import pytest
import b3.hexdump

from c3.constants import *
from c3.errors import *
from c3.signverify import SignVerify
from c3 import structure
from c3 import textfiles


@pytest.fixture
def c3m():
    c3_obj = SignVerify()
    return c3_obj


# ======== Priv Key encrypt Tests ==================================================================

# bare (not encrypted) key roundtrip
def test_privkey_bare(c3m):
    bare_priv = b"hello world"
    priv_block_bytes = structure.make_priv_block(bare_priv, bare=True)
    privd = structure.load_priv_block(priv_block_bytes)
    assert privd.priv_type == PRIVTYPE_BARE
    assert privd.key_type == KEYTYPE_ECDSA_256P
    decrypted_priv = c3m.decrypt_private_key(privd)     # should pass bare through
    assert decrypted_priv == bare_priv

# encrypted key roundtrip using an environment variable password
def test_privkey_env_var(c3m):
    bare_priv = b"hello world"
    os.environ["C3_PASSWORD"] = "Password01!"
    epriv = c3m.encrypt_private_key(bare_priv)
    priv_block_bytes = structure.make_priv_block(epriv)   # bare=False is the default

    privd = structure.load_priv_block(priv_block_bytes)
    assert privd.priv_type == PRIVTYPE_PASS_PROTECT
    assert privd.key_type == KEYTYPE_ECDSA_256P
    decrypted_priv = c3m.decrypt_private_key(privd)
    assert decrypted_priv == bare_priv

# glitch a privkey byte to exercise the integrity check
def test_privkey_bare_integrity(c3m):
    bare_priv = b"hello world"
    priv_block_bytes = structure.make_priv_block(bare_priv, bare=True)
    priv_block_bytes = priv_block_bytes[:16] + b"a" + priv_block_bytes[17:]
    with pytest.raises(IntegrityError):
        structure.load_priv_block(priv_block_bytes)

# Run me from TOP top level with: "python -m c3.tests.test_c3"

def interactive_password_test():        # note: not a pytest test
    c3m = SignVerify()
    bare_priv = b"hello world"
    print("- Encrypt & pack - ")
    epriv = c3m.encrypt_private_key(bare_priv)
    priv_block_bytes = structure.make_priv_block(epriv)
    if not priv_block_bytes:        # or exception?
        print(" - user abort -")
        return
    print("- load & decrypt - ")
    print(repr(priv_block_bytes))

    pd = structure.load_priv_block(priv_block_bytes)
    de_priv = c3m.decrypt_private_key(pd)
    if de_priv == bare_priv:
        print("SUCCESS - roundtripped key matches original")
    else:
        print("FAIL - roundtripped key does not match original")



# ======== Verify Tests ============================================================================

# Made with:  python commandline.py make --name="root1" --using=self --expiry=2022-09-09
root1 = """
2UK6AelNtgEJAGUJAAVyb290MRkBBXJvb3QxOQIBAQkDQMg/+gyc/F3JZs4rY6ya5d9cPdWd6Sjs
OD3UsENAitTCGrY0fceRKrndJKyXsVwOhwe2lUpmidVytXzAundMMWmpBAQJCcwfqQUEFgrMHwkB
SwkAQOe/rz9mLoBLLRCpFEr+emsHEvcH73vZislzpwqBHxD+tV23QFmTMjsJzKsPOQmmYOxMkH/U
EbKFB4gzSK0aT/EJAQVyb290MQ==
"""
root1_block = base64.b64decode(root1)

plaintext_payload = b"Hello this is the actual payload\n"

payload_and_inter2_wanted_name_root1 = """
2TerAulNbAkAIUhlbGxvIHRoaXMgaXMgdGhlIGFjdHVhbCBwYXlsb2FkCgkBRQkAQH510LkxZEZJ
Az8n/j0GRhSdyrnvudPhZ08GXUkltKZUIP2cMiNfQhiuLDvgbVho5zJZVrIrdSQEJU0xbYsZLWIB
AelNuAEJAGcJAAZpbnRlcjIZAQZpbnRlcjI5AgEBCQNAJnKBsUN6eOud943dJotzVG5RJtxCBfMr
ZUb7lPv9c1dxoMEFWQHBdotULByW1RtZkCZNE5pAx3Ll3gOyBD4keakEBBcJzB+pBQQWCswfCQFL
CQBA2w6v9I6uaICQ5V1d4R8Tsui/pLq45V2Rg7jNx1mypYXHDaQzpvrCbf/lR2Z9OgpTH8vmBafy
DYlgh26lpIO5HAkBBXJvb3Qx
"""
public_part = base64.b64decode(payload_and_inter2_wanted_name_root1)

def test_add_trusted_root1(c3m):
    c3m.add_trusted_certs(root1_block)


# Happy path
def test_verify_success_ext_root1(c3m):
    c3m.add_trusted_certs(root1_block)
    ret = c3m.verify(structure.load_pub_block(public_part))
    assert ret is True

# payload extractor & meta-data chain maker

def test_get_meta_root1(c3m):
    chain = structure.load_pub_block(root1_block)
    assert chain[0].cert.cert_id == b"root1"
    assert chain[0].sig.signing_cert_id == b"root1"
    # Meta is just chain without all the big-bytesy things so it's nicer to pprint()
    meta = structure.get_meta(chain)
    assert meta[0].cert.cert_id == b"root1"
    assert meta[0].sig.signing_cert_id == b"root1"

def test_get_meta_payload(c3m):
    chain = structure.load_pub_block(public_part)
    meta = structure.get_meta(chain)
    assert len(meta) == 1
    assert meta[0].cert.cert_id == b"inter2"
    assert meta[0].sig.signing_cert_id == b"root1"

def test_get_payload_root1(c3m):
    chain = structure.load_pub_block(root1_block)
    payload = structure.get_payload(chain)
    assert payload == b""

def test_get_payload_payload(c3m):
    chain = structure.load_pub_block(public_part)
    payload = structure.get_payload(chain)
    assert payload == plaintext_payload





# Glitch the payload contents so the signature fails to verify
def test_verify_signature_fail(c3m):
    public_part_glitched = public_part[:100] + b"X" + public_part[101:]
    with pytest.raises(InvalidSignatureError):
        c3m.verify(structure.load_pub_block(public_part_glitched))


# Apart from actual signature fails, there are 3 ways for this to fail:
# 1) "fell off" - unnamed issuer cert and no next cert in line ("short chain")
# 2) Cant Find Named Cert - in the cert store / trust store / certs_by_name etc
# 3) Last cert is self-signed and verified OK but isn't in the trust store.

# cut inter2 off the end of payload_and_chain_with_wanted_name
# to trigger "next cert is the signer but there is no next cert" failure mode
# Don't need root1 loaded because it doesn't get that far
def test_verify_short_chain(c3m):
    public_part_without_inter2 = public_part[:115]
    with pytest.raises(ShortChainError):
        c3m.verify(structure.load_pub_block(public_part_without_inter2))


# Without loading root1 to trusted store first
def test_verify_cert_not_found_error(c3m):
    with pytest.raises(CertNotFoundError):
        c3m.verify(structure.load_pub_block(public_part))



payload_and_chain_with_root_selfsigned_included = """
2TffA+lNbAkAIUhlbGxvIHRoaXMgaXMgdGhlIGFjdHVhbCBwYXlsb2FkCgkBRQkAQEm0pCEzDFCz
ZQ30w/HbGOXAO+VA9h/EzUdE1lZCqpOnRF3yDwjyXukQvUP8nAhBN7u/gRzy0lkS80BqjpFQ74kB
AelNsgEJAGcJAAZpbnRlcjMZAQZpbnRlcjM5AgEBCQNAYL5PIvF/unXMHSQrqtmdYxMS0R2gv5mg
AdViuQxCA8mTU2hlvqLrdlNuFFxNdOQ77z03EOwpyJMeD3TJ8kJM0akEBBcJzB+pBQQWCswfCQFF
CQBAgKHkaHcq4cDgpVhOccTOqouUKCZtF9gbaAESuHu2E7Oyj0rjXBVTiiBqw9zsOi8Gw7JiGWwJ
TiyR0qThWHHowQEB6U22AQkAZQkABXJvb3QxGQEFcm9vdDE5AgEBCQNAyD/6DJz8XclmzitjrJrl
31w91Z3pKOw4PdSwQ0CK1MIatjR9x5Equd0krJexXA6HB7aVSmaJ1XK1fMC6d0wxaakEBAkJzB+p
BQQWCswfCQFLCQBA57+vP2YugEstEKkUSv56awcS9wfve9mKyXOnCoEfEP61XbdAWZMyOwnMqw85
CaZg7EyQf9QRsoUHiDNIrRpP8QkBBXJvb3Qx
"""

# a fully valid chain with a selfsign at the end, should still fail with UntrustedChainError
def test_verify_untrusted_chain(c3m):
    public_part_selfsign_incl = base64.b64decode(payload_and_chain_with_root_selfsigned_included)
    with pytest.raises(UntrustedChainError):
        c3m.verify(structure.load_pub_block(public_part_selfsign_incl))


# ---- Test load error handling ----

def test_load_empty(c3m):
    with pytest.raises(StructureError):
        structure.load_pub_block(b"")

def test_load_none(c3m):
    with pytest.raises(StructureError):
        structure.load_pub_block(None)

def test_load_nulls(c3m):
    with pytest.raises(StructureError):
        structure.load_pub_block(b"\x00\x00\x00\x00\x00\x00\x00\x00")


# ======== Friendly Fields Tests ===================================================================


def test_make_friendly_fields(c3m):
    pub_part = base64.b64decode(root1)
    lines_str = textfiles.make_friendly_fields(pub_part, CERT_SCHEMA, ["subject_name", "expiry_date"])
    # print(repr(lines_str))
    assert lines_str == '[ Subject Name ]  root1\n[ Expiry Date  ]  9 September 2022'



ff_root1_pub_text = """
--------------------[ root1 - Payload & Public Certs ]----------------------
[ Subject Name ]  root1
[ Expiry Date  ]  9 September 2022
2UK6AelNtgEJAGUJAAVyb290MRkBBXJvb3QxOQIBAQkDQMg/+gyc/F3JZs4rY6ya5d9cPdWd6Sjs
OD3UsENAitTCGrY0fceRKrndJKyXsVwOhwe2lUpmidVytXzAundMMWmpBAQJCcwfqQUEFgrMHwkB
SwkAQOe/rz9mLoBLLRCpFEr+emsHEvcH73vZislzpwqBHxD+tV23QFmTMjsJzKsPOQmmYOxMkH/U
EbKFB4gzSK0aT/EJAQVyb290MQ==

this_part_should_be_ignored
"""

ff_root1_b64_block = base64.b64decode('\n'.join(ff_root1_pub_text.splitlines()[4:8]))

def test_ff_check_happy_path(c3m):
    ret = textfiles.check_friendly_fields(ff_root1_pub_text, CERT_SCHEMA)
    assert ret == ff_root1_b64_block

def test_ff_busted_vertical_1(c3m):
    busted_vertical_structure = "\n\n".join(ff_root1_pub_text.splitlines())
    with pytest.raises(StructureError, match="structure is invalid"):
        textfiles.check_friendly_fields(busted_vertical_structure, CERT_SCHEMA)

def test_ff_bad_field(c3m):
    bad_ff = ff_root1_pub_text.replace("Date  ]", "Date]")
    with pytest.raises(TamperError, match="format for visible"):
        textfiles.check_friendly_fields(bad_ff, CERT_SCHEMA)

def test_ff_spurious_field(c3m):
    spurious_field = "[ Spurious Field ]  Hello world"
    bad_ff = ff_root1_pub_text.replace("[ Subject Name ]  root1", spurious_field)
    with pytest.raises(TamperError, match="not present in the secure area"):
        textfiles.check_friendly_fields(bad_ff, CERT_SCHEMA)

def test_ff_field_value_mismatch(c3m):
    spurious_field = "[ Subject Name ]  Harold"
    bad_ff = ff_root1_pub_text.replace("[ Subject Name ]  root1", spurious_field)
    with pytest.raises(TamperError, match="does not match secure"):
        textfiles.check_friendly_fields(bad_ff, CERT_SCHEMA)



# todo: friendly fields tests.





# ======== Sign Tests ==============================================================================

#               |     no payload                  payload
#  -------------+----------------------------------------------------------
#  using cert   |     make chain signer           sign payload
#               |
#  using self   |     make self signer            ERROR invalid state


#               |     using_name                  no using_name
#  -------------+----------------------------------------------------------
#  using_pub    |     invalid                     append cert, blank name
#               |
#  no using_pub |     no append cert, link name   invalid


def test_make_selfsigned(c3m):
    # make a selfsigned then verify it and check the cert name == the sig name
    expiry = datetime.date(2023, 9, 9)

    pub_part_bytes, priv_part_bytes = c3m.make_sign(action=MAKE_SELFSIGNED, name=u"test1", expiry=expiry)
    c3m.add_trusted_certs(pub_part_bytes)

    chain = structure.load_pub_block(pub_part_bytes)
    ret = c3m.verify(chain)
    assert ret is True      # no payload, successful verify
    assert chain[0].cert.cert_id == chain[0].sig.signing_cert_id    # self-signed


def test_make_supply_neither_inval(c3m):
    with pytest.raises(ValueError):
        inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9")


def test_make_supply_both_inval(c3m):
    with pytest.raises(ValueError):
        inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_pub=b"a", using_name="root9")
        # Note it doen't get to needing the missing using_priv or expiry


def test_make_inter_name(c3m):
    expir = datetime.date(2023, 9, 9)
    # Root cert
    root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root9", expiry=expir)
    c3m.add_trusted_certs(root_pub)

    inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_priv=root_priv, expiry=expir,
                                          using_pub=root_pub, using_name="root9", link=LINK_NAME)

    chain = structure.load_pub_block(inter_pub)
    ret = c3m.verify(chain)
    assert ret is True      # no payload, successful verify


def test_make_inter_append(c3m):
    expir = datetime.date(2023, 9, 9)
    # Root cert
    root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root9", expiry=expir)
    c3m.add_trusted_certs(root_pub)

    inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_priv=root_priv, expiry=expir,
                                          using_pub=root_pub, using_name="root9", link=LINK_APPEND)

    chain = structure.load_pub_block(inter_pub)
    ret = c3m.verify(chain)
    assert ret is True      # no payload, successful verify


def test_make_inter_append_expired_root(c3m):
    expir_root = datetime.date(2021, 9, 9)
    expir = datetime.date(2023, 9, 9)
    # Root cert
    root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root9", expiry=expir_root)
    c3m.add_trusted_certs(root_pub)
    with pytest.raises(CertExpired):
        c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_pub=root_pub, using_priv=root_priv, expiry=expir)


# Note that this doesn't fail, even though we are *appending* the root9 cert itself into the chain
#      which you're not supposed to do. It succeeds because root9 is in trusted_certs and verify
#      sees that the NAME root9 is in trusted_certs so sets the found_in_trusted flag so that
#      UntrustedChainError doesn't trigger at the end.

# Note this looks like it would open us up to malicious actors appending their own cert with the same
#      NAME, but the actual signature verification step is always done, which defends against this,
#      as shown by the next test.


def test_sign_rootcert_namecollide(c3m):
    expir = datetime.date(2023, 9, 9)
    # Legit guy
    root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root5", expiry=expir)
    c3m.add_trusted_certs(root_pub)
    # Attacker guy
    evil_pub, evil_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root5", expiry=expir)   # NOTE same name
    # evil chain
    inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_pub=evil_pub, using_priv=evil_priv, expiry=expir)
    chain = structure.load_pub_block(inter_pub)
    with pytest.raises(InvalidSignatureError):
        ret = c3m.verify(chain)



def test_sign_payload(c3m):
    expir = datetime.date(2023, 9, 9)
    root_pub, root_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root9", expiry=expir)

    inter_pub, inter_priv = c3m.make_sign(MAKE_INTERMEDIATE, name="inter9", using_pub=root_pub, using_name="root9", using_priv=root_priv, expiry=expir, link=LINK_NAME)

    payload = b"How are you gentlemen"
    signed_payload, should_be_none = c3m.make_sign(SIGN_PAYLOAD, payload=payload, using_pub=inter_pub, using_priv=inter_priv)
    assert should_be_none is None

    chain = structure.load_pub_block(signed_payload)
    c3m.add_trusted_certs(root_pub)
    verify_ok = c3m.verify(chain)
    assert verify_ok is True



def test_keypair_matching(c3m):
    expir = datetime.date(2023, 9, 9)
    r1_pub, r1_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root1", expiry=expir)
    r2_pub, r2_priv = c3m.make_sign(MAKE_SELFSIGNED, name="root1", expiry=expir)  # note simulating same name error

    with pytest.raises(SignError, match="key do not match"):
        c3m.make_sign(MAKE_INTERMEDIATE, name="inter2", using_pub=r1_pub,
                      using_name="root1", using_priv=r2_priv, expiry=expir,
                      link=LINK_NAME)


# ---- Truncate and glitch loops -----

# Testing what happens if the public_part buffer is incomplete
# (And finding out exactly where to truncate public_part for the short-chain test above)

def truncmain():
    c3m = SignVerify()
    c3m.add_trusted_certs(root1_block)
    buf = public_part[:]

    for i in range(len(buf)+1, 1, -1):
        buf2 = buf[:i]
        try:
            xx = c3m.load_pub_block(buf2)
        except Exception as e:
            # print("%4i    load   %20s" % (i,e))
            continue
        try:
            c3m.verify(xx)
        except Exception as e:
            print("%4i  verify   %r" % (i,e))
            continue
        print("%4i   - SUCCESS -" % (i,))


# glitch a byte anywhere? in the chain to trigger signature fails.

def glitchmain():
    c3m = SignVerify()
    c3m.add_trusted_certs(root1_block)
    buf = public_part[:]

    for i in range(len(buf)):
        buf2 = buf[:i] + b"\x00" + buf[i+1:]
        try:
            xx = c3m.load_pub_block(buf2)
        except Exception as e:
            print("%4i    load   %20s" % (i,e))
            if "index out of" in str(e):
                print()
                print(traceback.format_exc())
                print()
            continue
        try:
            c3m.verify(xx)
        except Exception as e:
            print("%4i  verify   %r" % (i,e))
            continue
        print("%4i   - SUCCESS -" % (i,))

def smallrandfuzz():
    c3m = SignVerify()
    z = {}
    i = 0
    while True:
        i += 1
        buf = random.randbytes(40)
        #buf = b"\xdd\x37\x40\xed\x4d\x30\x44" + random.randbytes(60)
        try:
            xx = c3m.load_pub_block(buf)
            out = "omg SUCCESS omg"
        except Exception as e:
            out = str(e)
        z[out] = z.get(out,0) + 1

        if i % 100000 == 0:
            print()
            pprint(z)
            return

# glitch a byte in the privkey block processing to ensure the decode integrity checks dont fail

def bare_glitch_loop():
    c3m = SignVerify()
    bare_priv = b"hello world"
    priv_block_bytes = c3m.make_encrypt_private_key_block(bare_priv, bare=True)

    print("=== Known-good ===")
    print(b3.hexdump.hexdump(priv_block_bytes))
    pd = c3m.load_priv_block(priv_block_bytes)
    pprint(pd)
    print()

    for i in range(len(priv_block_bytes)):
        buf = priv_block_bytes[:i] + b"\x0f" + priv_block_bytes[i+1:]
        try:
            pd = c3m.load_priv_block(buf)
            print(i, " - Success -")
            #print(b3.hexdump.hexdump(buf))
            #pprint(pd)
            #print()
        except Exception as e:
            print(i, str(e))




if __name__ == '__main__':
    #truncmain()
    #glitchmain()
    #smallrandfuzz()
    #bare_glitch_loop()
    interactive_password_test()



# ---- Basic fuzzing of the initial header check ----
#
# def FuzzEKH():
#     for i in range(0,255):
#         buf = six.int2byte(i) #+ b"\x0f\x55\x55"
#         try:
#             ppkey, index = expect_key_header([KEY_LIST_PAYLOAD, KEY_LIST_CERTS], b3.LIST, buf, 0)
#             print("%4i %02x - SUCCESS - key = %r" % (i,i, ppkey))
#         except Exception as e:
#             print("%4i %02x -  %s" % (i,i, e))
#             #print(traceback.format_exc())
#
# def FuzzEKH2():
#     i = 0
#     z = {}
#     while True:
#         i += 1
#         buf = random.randbytes(20)
#         try:
#             ppkey, index = expect_key_header([KEY_LIST_PAYLOAD, KEY_LIST_CERTS], b3.LIST, buf, 0)
#             out = "SUCCESS - key = %r" % ppkey
#         except Exception as e:
#             out = "%r" % e
#
#         #print(out)
#         z[out] = z.get(out,0) + 1
#
#         if i % 100000 == 0:
#             print()
#             print(len(z))
#             pprint(z)




# C3 private & public binary block structure saving, loading & validation

import copy, binascii, datetime

from c3.constants import *
from c3.errors import StructureError, IntegrityError, CertExpired


class AttrDict(dict):
    def __getattr__(self, name):
        return self[name]
    def __deepcopy__(self, memo):
        return self.__class__({k: copy.deepcopy(v, memo) for k, v in self.items()})


# --- Private part/block save & load ---

# in:  private key bytes, and if they are unencrypted (bare) or not
# out: block bytes with header.

def make_priv_block(priv_bytes, bare=False):
    privd = AttrDict()
    privd["key_type"] = KEYTYPE_ECDSA_256P
    privd["priv_type"] = PRIVTYPE_BARE if bare else PRIVTYPE_PASS_PROTECT
    privd["priv_data"] = priv_bytes
    privd["crc32"] = binascii.crc32(privd.priv_data, 0) % (1 << 32)
    out_bytes = b3.schema_pack(PRIV_CRCWRAPPED, privd)
    out_bytes_with_hdr = b3.encode_item_joined(KEY_PRIV_CRCWRAPPED, b3.DICT, out_bytes)
    return out_bytes_with_hdr


# in: block bytes from e.g. LoadFiles
# out: DICT with private key + metadata
# sanity & crc32 check the priv block, then shuck it and return the inner data.
# Caller must then decrypt the private key if needed.

def load_priv_block(block_bytes):
    _, index = expect_key_header([KEY_PRIV_CRCWRAPPED], b3.DICT, block_bytes, 0)
    privd = AttrDict(b3.schema_unpack(PRIV_CRCWRAPPED, block_bytes[index:]))
    # --- Sanity checks ---
    schema_ensure_mandatory_fields(PRIV_CRCWRAPPED, privd)
    if privd.priv_type not in KNOWN_PRIVTYPES:
        raise StructureError("Unknown privtype %d in priv block (wanted %r)" % (privd.priv_type, KNOWN_PRIVTYPES))
    if privd.key_type not in KNOWN_KEYTYPES:
        raise StructureError("Unknown keytype %d in priv block (wanted %r)" % (privd.key_type, KNOWN_KEYTYPES))
    # --- Integrity check ---
    data_crc = binascii.crc32(privd.priv_data, 0) % (1 << 32)
    if data_crc != privd.crc32:
        raise IntegrityError("Private key block failed data integrity check (crc32)")
    return privd


# --- Public part loader ---

# In: public_part bytes
# Out: data-and-sig list cert chain structure
# Note: the inverse of this function is part of make_sign

def load_pub_block(public_part):
    # The public part should have an initial header that indicates whether the first das is a payload or a cert
    ppkey, index = expect_key_header([KEY_LIST_PAYLOAD, KEY_LIST_CERTS], b3.LIST, public_part, 0)
    public_part = public_part[index:]               # chop off the header

    # Should be a list of DAS structures, so pythonize the list
    chain = list_of_schema_unpack(DATA_AND_SIG, [KEY_DAS], public_part)

    # unpack the certs & sigs in chain
    for i, das in enumerate(chain):
        # dont unpack cert if this is the first das and ppkey is PAYLOAD
        if i > 0 or ppkey == KEY_LIST_CERTS:
            das["cert"] = AttrDict(b3.schema_unpack(CERT_SCHEMA, das.data_part))
            schema_ensure_mandatory_fields(CERT_SCHEMA, das.cert)

        das["sig"] = AttrDict(b3.schema_unpack(SIG_SCHEMA, das.sig_part))
        schema_ensure_mandatory_fields(SIG_SCHEMA, das.sig)

    return chain


# --- Structure helper functions ---

# Expects a list items which are the same schema object. This should eventually be part of b3.

def list_of_schema_unpack(schema, want_keys, buf):
    end = len(buf)
    index = 0
    out = []
    while index < end:
        try:
            key, data_type, has_data, is_null, data_len, index = b3.item.decode_header(buf, index)
        except (IndexError, UnicodeDecodeError):
            raise StructureError("List item header structure is invalid")
        if key not in want_keys:
            raise StructureError \
                ("List item header key invalid - wanted %r got %r" % (want_keys, key))
        if data_type != b3.DICT:
            raise StructureError("List item header type invalid - wanted DICT got %r" % data_type)
        if not has_data or data_len == 0:
            raise StructureError("List item header invalid - no data")

        das_bytes = b3.decode_value(data_type, has_data, is_null, data_len, buf, index)

        if len(das_bytes) == 0:
            raise StructureError("List item data is missing")

        # Now unpack the actual dict too
        dx = b3.schema_unpack(schema, das_bytes)
        schema_ensure_mandatory_fields(schema, dx)
        out.append(AttrDict(dx))
        index += data_len
    return out


def schema_ensure_mandatory_fields(schema, dx):
    for field_def in schema:                    # by name
        # only check if mandatory bool flag is both present AND true.
        if len(field_def) > 3 and field_def[3] is True:
            field_name = field_def[1]
            if field_name not in dx:
                raise StructureError("Required schema field '%s' is missing" % field_name)
            if not dx[field_name]:
                raise StructureError("Mandatory field '%s' is %r" % (field_name, dx[field_name]))



# Index and Unicode are the only two unhandled exception types that b3's decode_header code produces when fuzzed.
# IndexError trying to decode a bad varint for ext_type, datalen or number key.
# Unicode for when b3 thinks there's a utf8 key but the utf8 is bad.
def expect_key_header(want_keys, want_type, buf, index):
    if not buf:
        raise StructureError("No data - buffer is empty or None")
    try:
        key, data_type, has_data, is_null, data_len, index = b3.decode_header(buf, index)
    except (IndexError, UnicodeDecodeError):
        raise StructureError("Header structure is invalid")  # from None
        # raise .. from None disables py3's chaining (cleaner unhandled prints) but isnt legal py2
    if key not in want_keys:
        raise StructureError \
            ("Incorrect key in header - wanted %r got %s" % (want_keys, repr(key)[:32]))
    if data_type != want_type:
        raise StructureError \
            ("Incorrect type in header - wanted %r got %s" % (want_type, repr(data_type)[:32]))
    if not has_data:
        raise StructureError("Invalid header - no has_data")
    if index == len(buf):
        raise StructureError("No data after header - buffer is empty")
    return key, index


# This does not ensure mandatory fields are present like load_pub_block() does, so it can be used for
# more things e.g. friendly_fields and check_expiry.

def extract_first_dict(part_block, schema):
    if schema == CERT_SCHEMA:       # public part block
        ppkey, index = expect_key_header([KEY_LIST_PAYLOAD, KEY_LIST_CERTS], b3.LIST, part_block, 0)
        public_part = part_block[index:]
        das0 = list_of_schema_unpack(DATA_AND_SIG, [KEY_DAS], public_part)[0]
        dx0 = AttrDict(b3.schema_unpack(schema, das0.data_part))
    elif schema == PRIV_CRCWRAPPED:    # private part block
        ppkey, index = expect_key_header([KEY_PRIV_CRCWRAPPED,], b3.DICT, part_block, 0)
        private_part = part_block[index:]
        dx0 = AttrDict(b3.schema_unpack(schema, private_part))
    else:
        raise TypeError("Unknown schema for first-dict extract")
    return dx0

def ensure_not_expired(using_pub):
    dx0 = extract_first_dict(using_pub, CERT_SCHEMA)
    expiry = dx0["expiry_date"]
    if datetime.date.today() > expiry:
        raise CertExpired("cert specified by --using has expired")
    return True



# --- Output/Results fetchers ---

# In: chain from load()
# Out: payload bytes

def get_payload(chain):
    first_das = chain[0]
    if "cert" not in first_das:         # first data_and_sig element's data_part is a payload
        return first_das.data_part
    return b""      # first data_and_sig element's data_part is a cert


# Delivering a cut-down Chain for meta:
# chain is list of:
#   data_part,  ->  cert (or payload)   ->  cert fields
#   sig_part,   ->  sig                 ->  signature & signing_cert_id

# In: chain from load()
# Out: trimmed-down metadata-only chain

def get_meta(chain):
    st = 0
    first_das = chain[0]
    if "cert" not in first_das:     # skip the first data_and_sig if it's a payload one
        st = 1
    chain2 = copy.deepcopy(chain[st:])
    for i in chain2:
        del i["data_part"]
        del i["sig_part"]
        del i["cert"]["public_key"]
        del i["sig"]["signature"]
    return chain2


# For error message readability
def ctnm(das):
    if not das:
        return ""
    if "cert" in das:
        return " (cert %r) " % das.cert.cert_id
    else:
        return " (payload) "

# --------------------------------------------------------------------------------------------------


# Making ULIDS

# Courtesy of https://github.com/valohai/ulid2/blob/master/ulid2/__init__.py
#
# import time, calendar, struct
# _last_entropy = None
# _last_timestamp = None
#
# def generate_binary_ulid(timestamp=None, monotonic=False):
#     """
#     Generate the bytes for an ULID.
#     :param timestamp: An optional timestamp override.
#                       If `None`, the current time is used.
#     :type timestamp: int|float|datetime.datetime|None
#     :param monotonic: Attempt to ensure ULIDs are monotonically increasing.
#                       Monotonic behavior is not guaranteed when used from multiple threads.
#     :type monotonic: bool
#     :return: Bytestring of length 16.
#     :rtype: bytes
#     """
#     global _last_entropy, _last_timestamp
#     if timestamp is None:
#         timestamp = time.time()
#     elif isinstance(timestamp, datetime.datetime):
#         timestamp = calendar.timegm(timestamp.utctimetuple())
#
#     ts = int(timestamp * 1000.0)
#     ts_bytes = struct.pack(b'!Q', ts)[2:]
#     entropy = os.urandom(10)
#     if monotonic and _last_timestamp == ts and _last_entropy is not None:
#         while entropy < _last_entropy:
#             entropy = os.urandom(10)
#     _last_entropy = entropy
#     _last_timestamp = ts
#     return ts_bytes + entropy
#






# C3 Command-line operations (make keys, sign etc).
# Doubles as usage examples for the various components.

from __future__ import print_function

import sys, re, datetime
from pprint import pprint

from c3.constants import *
from c3 import signverify
from c3 import structure
from c3 import textfiles


# Policy: Be as janky with this stuff as we want right now, we're not doing a public release,
#         just making the UX not suck too bad for us. This means click is out, --nopassword=yes is ok.
# Policy: For simplicity for now, the subject names and cert_ids are the same. Later there should be ULIDs.

# - Turn into Package.
# * Clean up password prompts and commandline UX
# * improve expiry date parsing
# * Do build to pypi, 0.9.0

# Use cases:
# * Make license (sign),  verify license
# * Make acmd key, make (sign) acmd message, verify acmd message
# * make build key [nocode], sign build manifest [nocode], verify build manifest  [signverify]


# need to do __init__.py  and have an external test code.
# For signing things, and verifying things.


# [DO THIS DURING LICENSING]  - so e.g. licensing gets its own friendly fields.


#               |     no payload             payload
#  -------------+-------------------------------------------------
#  using cert   |     make chain signer      sign payload
#               |
#  using self   |     make self signer       ERROR invalid state


def CommandlineMain(zargs=None):
    print("Checking Function args received (for entry point testing): ",repr(zargs))
    if zargs is None:
        zargs = sys.argv[1:]
    if len(zargs) < 1:
        UsageBail()
    cmd = zargs[0].lower()
    args = ArgvArgs(zargs)

    c3m = signverify.SignVerify()

    # python commandline.py  make --name=root1 --using=self  --parts=split --expiry=2022-02-02
    # python commandline.py  make --name=inter1 --using=root1 --link=name --parts=combine --expiry=2022-02-02

    if cmd == "make":
        if "using" not in args:
            print("'make' needs --using=<name> or --using=self, please supply")
            return

        expiry = ParseExpiryDate(args.expiry)
        if args.using == "self":
            pub_block, priv = c3m.make_sign(action=MAKE_SELFSIGNED, name=args.name, expiry=expiry)
        else:
            if "link" not in args:
                print("'make' needs --link=append or --link=name, please supply")
                return

            upub, uepriv = textfiles.load_files(args.using)         # uses files
            upriv = c3m.decrypt_private_key(structure.load_priv_block(uepriv))  # (might) ask user for password
            link = {"append" : LINK_APPEND, "name" : LINK_NAME}[args.link]

            pub_block, priv = c3m.make_sign(action=MAKE_INTERMEDIATE, name=args.name, expiry=expiry,
                                 using_priv=upriv, using_pub=upub, using_name=args.using, link=link)

        bare = "nopassword" in args  # Note: has to be --nopassword=blah for now.
        if not bare:
            print("Setting password on private key-")
            epriv = c3m.encrypt_private_key(priv)
        else:
            epriv = priv
        epriv_block = structure.make_priv_block(epriv, bare)

        combine = True
        if "parts" in args and args.parts == "split":
            combine = False

        pub_ff_names = ["subject_name", "expiry_date", "issued_date"]
        pub_ffields = textfiles.make_friendly_fields(pub_block, CERT_SCHEMA, pub_ff_names)

        textfiles.write_files(args.name, pub_block, epriv_block, combine, pub_ff_lines=pub_ffields)
        return

    # python commandline.py  sign --payload=payload.txt --link=append  --using=inter1

    if cmd == "sign":
        if "payload" not in args:
            print("please supply --payload=<filename>")
            return
        payload_bytes = open(args.payload, "rb").read()

        upub, uepriv = textfiles.load_files(args.using)  # uses files
        upriv = c3m.decrypt_private_key(structure.load_priv_block(uepriv))  # (might) ask user for password
        link = {"append": LINK_APPEND, "name": LINK_NAME}[args.link]

        pub, priv = c3m.make_sign(action=SIGN_PAYLOAD, name=args.name, payload=payload_bytes,
                                  using_priv=upriv, using_pub=upub, link=link)

        # pub_ff_names = ["whatever", "app_specific", "fields_app_schema_has"]
        # pub_ffields = c3m.make_friendly_fields(pub, APP_SCHEMA, pub_ff_names)
        textfiles.write_files(args.payload, pub, b"", combine=False)   #, pub_ff_lines=pub_ffields))
        # Note: ^^ no private part, so no combine.         ^^^ how to friendly-fields for app
        return

    # python commandline.py  verify --name=payload.txt --trusted=root1

    if cmd == "verify":
        if "trusted" in args:
            print("Loading trusted cert ", args.trusted)
            tr_pub, _ = textfiles.load_files(args.trusted)
            c3m.add_trusted_certs(tr_pub)
        else:
            print("Please specify a trusted cert with --trusted=")
            return

        public_part, _ = textfiles.load_files(args.name)
        chain = structure.load_pub_block(public_part)
        ret = c3m.verify(chain)
        print("\n\nverify returns", repr(ret))
        if not ret:
            return
        print("Chain:")
        pprint(structure.get_meta(chain))
        print("Payload:")
        print(structure.get_payload(chain))
        return

    UsageBail("Unknown command")



def ParseExpiryDate(exp_str):
    exp_d = datetime.datetime.strptime(exp_str, "%Y-%m-%d").date()
    # Put some intelligent regex matches here so we can decode a few of the most common date formats
    # yes we need this.
    return exp_d


def UsageBail(msg=""):
    help_txt = """
    %s
    Usage:
    # python commandline.py  make --name=root1 --using=self  --parts=split
    # python commandline.py  make --name=inter1 --using=root1 --link=name --parts=combine   

    """ % (msg+"\n",)
    print(help_txt)
    sys.exit(1)


def ArgvArgs(zargs):
    args = structure.AttrDict()
    for arg in zargs:
        z = re.match(r"^--(\w+)=(.+)$", arg)
        if z:
            k, v = z.groups()
            args[k] = v
    return args

if __name__ == "__main__":
    CommandlineMain()






# C3 private & public text-file format saving, loading & validation

import os, base64, re, functools, datetime

import b3

from c3 import structure
from c3.constants import CERT_SCHEMA, PRIV_CRCWRAPPED
from c3.errors import StructureError, TamperError

# ============================== File Saving/Loading ===========================================

# Policy: look for name.PRIVATE and name.PUBLIC (.b64.txt)
# Policy: split trumps combined.
# Policy: Return "" for private_part if there is none, callers can validate

def asc_header(msg):
    m2 = "[ %s ]" % msg
    offs = 37 - len(m2) // 2
    line = "-" * offs
    line += m2
    line += "-" * (76 - len(line))
    return line


def write_files(name, public_part, private_part=b"", combine=True, desc="", pub_ff_lines="", priv_ff_lines=""):
    pub_desc = desc if desc else (name + " - Payload & Public Certs")
    priv_desc = (desc or name) + " - PRIVATE Key"
    if pub_ff_lines:
        pub_ff_lines += "\n"
    pub_str = asc_header(pub_desc) + "\n" + pub_ff_lines + base64.encodebytes \
        (public_part).decode()
    if priv_ff_lines:
        priv_ff_lines += "\n"
    priv_str = asc_header(priv_desc) + "\n" + priv_ff_lines + base64.encodebytes \
        (private_part).decode()

    print()
    print(pub_str)
    print()
    print(priv_str)
    if combine:
        fname = name + ".b64.txt"
        with open(fname, "w") as f:
            f.write("\n " +pub_str)
            f.write("\n")
            f.write(priv_str + "\n")
        print("Wrote combined file: " ,fname)
    else:
        fname = name + ".public.b64.txt"
        with open(fname, "w") as f:
            f.write("\n " + pub_str + "\n")
        print("Wrote public file:  ", fname)

        if not private_part:
            return

        fname = name + ".PRIVATE.b64.txt"
        with open(fname, "w") as f:
            f.write("\n " + priv_str + "\n")
        print("Wrote PRIVATE file: ", fname)



def load_files(name):
    header_rex = r"^-+\[ (.*?) \]-+$"
    pub_text_block = ""
    priv_text_block = ""

    combine_name = name + ".b64.txt"
    if os.path.isfile(combine_name):
        print("Loading combined file ", combine_name)
        both_strs = open(combine_name, "r").read()

        # regex cap the header lines
        hdrs = list(re.finditer(header_rex, both_strs, re.MULTILINE))
        if len(hdrs) != 2:
            print(" Warning: number of headers in combined file is not 2")

        # Structure: first header, first data, second header, second data, end of file
        # data offsets are start-of-first-header : start-of-second-header,
        # because check_friendly_fields wants to see the headers too if they are there.
        block0_text = both_strs[hdrs[0].start() : hdrs[1].start()]
        block1_text = both_strs[hdrs[1].start( ):]

        # normally the second block is the private block, but if a user has shuffled things around
        # we cater for that by checking which block has 'PRIVATE' in its header description
        if "PRIVATE" in hdrs[0].group(1):       # Private block comes first (not the normal case)
            pub_text_block, priv_text_block = block1_text, block0_text
        else:   # Otherwise assume the public block comes first.
            pub_text_block, priv_text_block = block0_text, block1_text

    # Enable more-specific files to override the combined file, if both exist

    pub_only_name = name + ".public.b64.txt"
    if os.path.isfile(pub_only_name):
        print("Loading public file ", pub_only_name)
        pub_text_block = open(pub_only_name, "r").read()
        hdrs = list(re.finditer(header_rex, pub_text_block, re.MULTILINE))
        if len(hdrs) != 1:
            print(" Warning: too %s headers in public file" % ("many" if len(hdrs ) >1 else "few"))

    priv_only_name = name + ".PRIVATE.b64.txt"
    if os.path.isfile(priv_only_name):
        print("Loading private file ", priv_only_name)
        priv_text_block = open(priv_only_name, "r").read()
        hdrs = list(re.finditer(header_rex, priv_text_block, re.MULTILINE))
        if len(hdrs) != 1:
            print(" Warning: too %s headers in public file" % ("many" if len(hdrs) > 1 else "few"))

    # Ensure friendly (visible) text-fields (if any) match the secure binary info.
    # This also extracts and converts the base64 secure block parts.
    pub_block = check_friendly_fields(pub_text_block, CERT_SCHEMA)
    priv_block = check_friendly_fields(priv_text_block, PRIV_CRCWRAPPED)

    return pub_block, priv_block

# Like load_files but if the public block part is a string. (e.g. cert stored in code)

def pub_block_from_string(pub_text_block):
    pub_block = check_friendly_fields(pub_text_block, CERT_SCHEMA)
    return pub_block



# ============================== Friendly Fields ===============================================

# In: block_part bytes, schema for first dict, field names to output in friendly format
# Out: field names & values as text lines (or exceptions)

def make_friendly_fields(block_part, schema, friendly_field_names):
    # --- get to that first dict ---
    # Assume standard pub_bytes structure (chain with header)
    # We can't use load() here because load() does mandatory schema checks and we
    dx0 = structure.extract_first_dict(block_part, schema)

    # --- Cross-check whether wanted fields exist (and map names to types) ---
    # This is because we're doing this with payloads as well as certs
    # The rest of the SignVerify system is fully payload-agnostic but we aren't.
    types_by_name = {}
    for typ, name in [i[:2] for i in schema]:
        if name in dx0 and name in friendly_field_names:
            types_by_name[name] = typ
    if not types_by_name:
        raise ValueError("No wanted friendly fields found in the secure block")
        # note: should this just be a warning & continue?

    # --- Convert wanted fields to a textual representation where possible ---
    # order by the friendly_field_names parameter
    line_items = []
    for name in friendly_field_names:
        if name not in types_by_name:
            continue
        fname = name.title().replace("_" ," ")
        typ = types_by_name[name]
        val = dx0[name]     # in
        fval = ""   # out
        # --- Value converters ---
        if typ in (b3.BYTES, b3.LIST, b3.DICT, 11, 12):  # cant be str-converted
            raise TypeError("Visible field '%s' cannot be text-converted (type %s), skipping" %
            (name, b3.b3_type_name(typ)))
        elif typ == b3.SCHED:
            fval = "%s, %s" % (val.strftime("%-I:%M%p").lower(), val.strftime("%-d %B %Y"))
        elif typ == b3.BASICDATE:
            fval = val.strftime("%-d %B %Y")
        else:
            fval = str(val)
        line_items.append((fname, fval))

    # --- Make stuff line up nicely ---
    longest_name_len = functools.reduce(max, [len(i[0]) for i in line_items], 0)
    lines = ["[ %s ]  %s" % (fname.ljust(longest_name_len), fval) for fname, fval in
             line_items]
    return '\n'.join(lines)


# Note: unlike make_friendly_fields, we raise exceptions when something is wrong
# In: text with header line, friendly-fields lines, followed by the base64 of the secure public or private block.
# Out: exceptions or True.
# We're pretty strict compared to make, any deviations at all will raise an exception.
#   This includes spurious fields, etc.
# We expect there to be an empty line (or end of file) after the base64 block, and NOT more stuff.
#   This means EOF immediately after the base64 block is an error.


def check_friendly_fields(text_part, schema):
    types_by_name = {i[1]: i[0] for i in schema}
    # --- Ensure vertical structure is legit ---
    # 1 or no header line (-), immediately followed by 0 or more FF lines ([),
    # immediately followd by base64 then a mandatory whitespace (e.g empty line).
    lines = text_part.splitlines()
    c0s = ''.join([line[0] if line else ' ' for line in lines]) + ' '
    X = re.match(r"^\s*(-?)(\[*)([a-zA-Z0-9/=+]+) ", c0s)
    if not X:
        raise StructureError("File text vertical structure is invalid")
    ff_lines = lines[X.start(2): X.end(2)]  # extract FF lines
    b64_lines = lines[X.start(3): X.end(3)]  # extract base64 lines
    b64_block = ''.join(b64_lines)
    bytes_part = base64.b64decode(b64_block)

    # --- get to that first dict in the secure block ---
    # Assume standard pub_bytes structure (chain with header)
    # Let these just exception out.
    dx0 = structure.extract_first_dict(bytes_part, schema)

    # --- Cross-check each Friendy Field line ---
    for ff in ff_lines:
        # --- Extract friendly name & value ---
        fX = re.match(r"^\[ (.*) ]  (.*)$", ff)
        if not fX:
            raise TamperError("Invalid format for visible field line %r" % ff[:32])
        fname, fval = fX.groups()

        # --- convert name ---
        # spaces aren't allowed in code schema field names because AttrDict, only underscores so
        # 1:1 conversion back is easy.
        name = fname.strip().lower().replace(" ", "_")
        fval = fval.strip()  # some converters are finicky about trailing spaces

        # --- Check name presence ---
        if name not in types_by_name:
            raise TamperError("Visible field '%s' is not present in the secure area" % (name,))
        typ = types_by_name[name]

        # --- convert value ---
        if typ == b3.UTF8:
            val = str(fval)  # actually the incoming text should already be utf8 anyway
        elif typ == b3.UVARINT:
            val = int(fval)
        elif typ == b3.BOOL:
            val = bool(fval.lower().strip() == "True")
        # elif typ == b3.SCHED:   # todo: this is the wrong way around
        #    val = "%s, %s" % (fval.strftime("%-I:%M%p").lower(), fval.strftime("%-d %B %Y"))
        elif typ == b3.BASICDATE:
            val = datetime.datetime.strptime(fval, "%d %B %Y").date()
        else:
            raise TamperError("Visible field '%s' cannot be type-converted" % (name,))

        # --- Compare value ---
        if name not in dx0:  # could happen if field is optional in the schema
            raise TamperError("Visible field '%s' is not present in the secure area" % (name,))
        secure_val = dx0[name]
        if secure_val != val:
            raise TamperError("Field '%s' visible value %r does not match secure value %r" % (
            name, val, secure_val))

    return bytes_part  # success

